# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from fabric.api import env
from dockermap.functional import lazy_once
# noinspection PyUnresolvedReferences
from dockermap.map.yaml import (yaml, load_file, load_map, load_map_file,
                                load_clients as _load_clients,
                                load_clients_file as _load_clients_file)
from .apiclient import DockerClientConfiguration


env_get = lambda v: env[v]


def expand_env_lazy(loader, node):
    """
    Substitutes a variable read from a YAML node with the value stored in Fabric's ``env`` dictionary. Creates an
    object for late resolution.

    :param loader: YAML loader.
    :type loader: yaml.loader.SafeLoader
    :param node: Document node.
    :type node: ScalarNode
    :return: Corresponding value stored in the ``env`` dictionary.
    :rtype: any
    """
    val = loader.construct_scalar(node)
    return lazy_once(env_get, val)


def expand_env(loader, node):
    """
    Substitutes a variable read from a YAML node with the value stored in Fabric's ``env`` dictionary.

    :param loader: YAML loader.
    :type loader: yaml.loader.SafeLoader
    :param node: Document node.
    :type node: ScalarNode
    :return: Corresponding value stored in the ``env`` dictionary.
    :rtype: any
    """
    val = loader.construct_scalar(node)
    return env[val]


def load_clients(stream):
    """
    Loads client configurations from a YAML document stream.

    :param stream: YAML stream.
    :type stream: file
    :return: A dictionary of client configuration objects.
    :rtype: dict[unicode, dockerfabric.apiclient.DockerClientConfiguration]
    """
    return _load_clients(stream, configuration_class=DockerClientConfiguration)


def load_clients_file(filename):
    """
    Loads client configurations from a YAML file.

    :param filename: YAML file name.
    :type filename: unicode
    :return: A dictionary of client configuration objects.
    :rtype: dict[unicode, dockerfabric.apiclient.DockerClientConfiguration]
    """
    return _load_clients_file(filename, configuration_class=DockerClientConfiguration)


yaml.add_constructor('!env_lazy', expand_env_lazy, yaml.SafeLoader)
yaml.add_constructor('!env', expand_env, yaml.SafeLoader)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric import operations
from fabric.context_managers import hide
from fabric.utils import error


def stdout_result(cmd, expected_errors=(), shell=True, sudo=False, quiet=False):
    """
    Runs a command and returns the result, that would be written to `stdout`, as a string. The output itself can
    be suppressed.

    :param cmd: Command to run.
    :type cmd: unicode
    :param expected_errors: If the return code is non-zero, but found in this tuple, it will be ignored. ``None`` is
      returned in this case.
    :type expected_errors: tuple
    :param shell: Use a shell.
    :type shell: bool
    :param sudo: Use `sudo`.
    :type sudo: bool
    :param quiet: If set to ``True``, does not show any output.
    :type quiet: bool
    :return: The result of the command as would be written to `stdout`.
    :rtype: unicode
    """
    which = operations.sudo if sudo else operations.run
    with hide('warnings'):
        result = which(cmd, shell=shell, quiet=quiet, warn_only=True)
    if result.return_code == 0:
        return result

    if result.return_code not in expected_errors:
        error("Received unexpected error code {0} while executing!".format(result.return_code))
    return None


def check_int(value):
    """
    Tests a given string for a possible conversion to integer. Uses Fabric's :func:`fabric.utils.error` instead of
    raising a :class:`TypeError`. ``None`` is not converted but returns ``None`` instead.

    :param value: Value to test for conversion.
    :type value: unicode
    :return: Integer value.
    :rtype: int
    """
    if value is not None:
        try:
            return int(value)
        except TypeError:
            error("Unexpected non-integer value '{0}'.".format(value))
    return None


single_line = lambda val: val.split('\n')[0] if val is not None else None


def single_line_stdout(cmd, expected_errors=(), shell=True, sudo=False, quiet=False):
    """
    Runs a command and returns the first line of the result, that would be written to `stdout`, as a string.
    The output itself can be suppressed.

    :param cmd: Command to run.
    :type cmd: unicode
    :param expected_errors: If the return code is non-zero, but found in this tuple, it will be ignored. ``None`` is
      returned in this case.
    :type expected_errors: tuple
    :param shell: Use a shell.
    :type shell: bool
    :param sudo: Use `sudo`.
    :type sudo: bool
    :param quiet: If set to ``True``, does not show any output.
    :type quiet: bool
    :return: The result of the command as would be written to `stdout`.
    :rtype: unicode
    """
    return single_line(stdout_result(cmd, expected_errors, shell, sudo, quiet))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import six

from fabric.api import env
from fabric.network import needs_host


@needs_host
def get_current_roles():
    """
    Determines the list of roles, that the current host is assigned to. If ``env.roledefs`` is not set, an empty list
    is returned.

    :return: List of roles of the current host.
    :rtype: list
    """
    current_host = env.host_string
    roledefs = env.get('roledefs')
    if roledefs:
        return [role for role, hosts in six.iteritems(roledefs) if current_host in hosts]
    return []


def get_role_addresses(role_name, interface_name):
    roledefs = env.get('roledefs')
    clients = env.get('docker_clients')
    if roledefs and clients:
        role_hosts = roledefs.get(role_name)
        if role_hosts:
            return set(client_config.interfaces[interface_name]
                       for client_name, client_config in six.iteritems(clients)
                       if client_config.get('fabric_host') in role_hosts)
    return set()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric.state import env
from fabric.utils import puts

from .base import ConnectionDict, get_local_port
from .tunnel import LocalTunnel


class SocketTunnels(ConnectionDict):
    """
    Cache for **socat** tunnels to the remote machine.

    Instantiation of :class:`SocketTunnel` can be configured with ``env.socat_quiet``, setting
    the ``quiet`` keyword argument.
    """
    def __getitem__(self, item):
        """
        :param item: Tuple of remote socket name, remote port, and local port number.
        :type item: tuple
        :return: Socket tunnel
        :rtype: SocketTunnel
        """
        def _connect_socket_tunnel():
            local_port = get_local_port(init_local_port)
            svc = SocketTunnel(remote_socket, local_port, env.get('socat_quiet', True))
            svc.connect()
            return svc

        remote_socket, init_local_port = item
        key = env.host_string, remote_socket
        return self.get_or_create_connection(key, _connect_socket_tunnel)


socat_tunnels = SocketTunnels()


class SocketTunnel(LocalTunnel):
    """
    Establish a tunnel from the local machine to the SSH host and from there start a **socat** process for forwarding
    traffic between the remote-end `stdout` and a Unix socket.

    :param remote_socket: Unix socket to connect to on the remote machine.
    :type remote_socket: unicode
    :param local_port: Local TCP port to use for the tunnel.
    :type local_port: int
    :param quiet: If set to ``False``, the **socat** command line on the SSH channel will be written to `stdout`.
    :type quiet: bool
    """
    def __init__(self, remote_socket, local_port, quiet=True):
        dest = 'STDIO'
        src = 'UNIX-CONNECT:{0}'.format(remote_socket)
        self.quiet = quiet
        self._socat_cmd = ' '.join(('socat', dest, src))
        super(SocketTunnel, self).__init__(local_port)

    def get_channel(self, transport, remote_addr, local_peer):
        channel = transport.open_channel('session')
        if channel is None:
            raise Exception("Failed to open channel on the SSH server.")
        if not self.quiet:
            puts(self._socat_cmd)
        channel.exec_command(self._socat_cmd)
        return channel



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric.api import env, sudo
from fabric.utils import puts, fastprint, error

from dockermap.client.base import LOG_PROGRESS_FORMAT, DockerStatusError
from dockermap.api import DockerClientWrapper
from .base import DockerConnectionDict, get_local_port, FabricClientConfiguration, FabricContainerClient
from .socat import socat_tunnels
from .tunnel import local_tunnels


DEFAULT_TCP_HOST = 'tcp://127.0.0.1'
DEFAULT_SOCKET = '/var/run/docker.sock'
progress_fmt = LOG_PROGRESS_FORMAT.format


def _get_port_number(expr, port_loc):
    try:
        return int(expr)
    except TypeError:
        raise ValueError("Missing or invalid {0} port ({1}).".format(port_loc, expr))


def _get_socat_tunnel(address, local_port):
    init_local_port = _get_port_number(local_port, 'local')
    tunnel_local_port = get_local_port(init_local_port)
    socat_tunnel = socat_tunnels[(address, tunnel_local_port)]
    return '{0}:{1}'.format(DEFAULT_TCP_HOST, socat_tunnel.bind_port), socat_tunnel


def _get_local_tunnel(address, remote_port, local_port):
    host_port = address.partition('/')[0]
    host, __, port = host_port.partition(':')
    service_remote_port = _get_port_number(port or remote_port, 'remote')
    init_local_port = _get_port_number(local_port or port or remote_port, 'local')
    local_tunnel = local_tunnels[(host, service_remote_port, 'localhost', init_local_port)]
    return '{0}:{1}'.format(DEFAULT_TCP_HOST, local_tunnel.bind_port), local_tunnel


def _get_connection_args(base_url, remote_port, local_port):
    if env.host_string:
        if base_url:
            proto_idx = base_url.find(':/')
            if proto_idx >= 0:
                proto = base_url[:proto_idx]
                address = base_url[proto_idx + 2:]
                if proto in ('http+unix', 'unix'):
                    if address[:3] == '//':
                        address = address[1:]
                    elif address[0] != '/':
                        address = ''.join(('/', address))
                    return _get_socat_tunnel(address, local_port)
                return _get_local_tunnel(address.lstrip('/'), remote_port, local_port)
            elif base_url[0] == '/':
                return _get_socat_tunnel(base_url, local_port)
            return _get_local_tunnel(base_url, remote_port, local_port)
        return _get_socat_tunnel(DEFAULT_SOCKET, local_port)
    return base_url, None


class DockerFabricClient(DockerClientWrapper):
    """
    Docker client for Fabric.

    For functional enhancements to :class:`docker.client.Client`, see
    :class:`~dockermap.client.base.DockerClientWrapper`. This implementation only adds the possibility to build a
    tunnel through the current SSH connection and adds Fabric-usual logging.

    If a unix socket is used, `socat` will be started on the remote side to redirect it to a TCP port.

    :param base_url: URL to connect to; if not set, will refer to ``env.docker_base_url`` or use ``None``, which by
     default attempts a connection on a Unix socket at ``/var/run/docker.sock``.
    :type base_url: unicode
    :param tls: Whether to use TLS on the connection to the Docker service.
    :type tls: bool
    :param version: API version; if not set, will try to use ``env.docker_api_version``; otherwise defaults to
     :const:`~docker.constants.DEFAULT_DOCKER_API_VERSION`.
    :type version: unicode
    :param timeout: Client timeout for Docker; if not set, will try to use ``env.docker_timeout``; otherwise defaults to
     :const:`~docker.constants.DEFAULT_TIMEOUT_SECONDS`.
    :type timeout: int
    :param tunnel_remote_port: Optional, port of the remote service; if port is included in ``base_url``, the latter
     is preferred. If not set, will try to use ``env.docker_tunnel_remote_port``; otherwise defaults to ``None``.
    :type tunnel_remote_port: int
    :param tunnel_local_port: Optional, for SSH tunneling: Port to open towards the local end for the tunnel; if not
     provided, will try to use ``env.docker_tunnel_local_port``; otherwise defaults to the value of
     ``tunnel_remote_port`` or ``None`` for direct connections without an SSH tunnel.
    :type tunnel_local_port: int
    :param kwargs: Additional kwargs for :class:`docker.client.Client`
    """
    def __init__(self, base_url=None, tls=None, version=None, timeout=None, tunnel_remote_port=None,
                 tunnel_local_port=None, **kwargs):
        url = base_url or env.get('docker_base_url')
        use_tls = tls or (tls is None and env.get('docker_tls', False))
        api_version = version or env.get('docker_api_version')
        client_timeout = timeout or env.get('docker_timeout')
        remote_port = tunnel_remote_port or env.get('docker_tunnel_remote_port')
        local_port = tunnel_local_port or env.get('docker_tunnel_local_port', remote_port)
        conn_url, self._tunnel = _get_connection_args(url, remote_port, local_port)
        super(DockerFabricClient, self).__init__(base_url=conn_url, version=api_version, timeout=client_timeout,
                                                 tls=use_tls, **kwargs)

    def push_log(self, info, level=None, *args, **kwargs):
        """
        Prints the log as usual for fabric output, enhanced with the prefix "docker".

        :param info: Log output.
        :type info: unicode
        :param level: Logging level. Has no effect here.
        :type level: int
        """
        if args:
            msg = info % args
        else:
            msg = info
        try:
            puts('docker: {0}'.format(msg))
        except UnicodeDecodeError:
            puts('docker: -- non-printable output --')

    def push_progress(self, status, object_id, progress):
        """
        Prints progress information.

        :param status: Status text.
        :type status: unicode
        :param object_id: Object that the progress is reported on.
        :type object_id: unicode
        :param progress: Progress bar.
        :type progress: unicode
        """
        fastprint(progress_fmt(status, object_id, progress), end='\n')

    def close(self):
        """
        Closes the connection and any tunnels created for it.
        """
        try:
            super(DockerFabricClient, self).close()
        finally:
            if self._tunnel is not None:
                self._tunnel.close()

    def build(self, tag, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.build` with additional logging.
        """
        kwargs['raise_on_error'] = True
        self.push_log("Building image '{0}'.".format(tag))
        try:
            return super(DockerFabricClient, self).build(tag, **kwargs)
        except DockerStatusError as e:
            error(e.message)

    def create_container(self, image, name=None, **kwargs):
        """
        Identical to :meth:`docker.api.container.ContainerApiMixin.create_container` with additional logging.
        """
        name_str = " '{0}'".format(name) if name else ""
        self.push_log("Creating container{0} from image '{1}'.".format(name_str, image))
        return super(DockerFabricClient, self).create_container(image, name=name, **kwargs)

    def copy_resource(self, container, resource, local_filename):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.copy_resource` with additional logging.
        """
        self.push_log("Receiving tarball for resource '{0}:{1}' and storing as {2}".format(container, resource, local_filename))
        super(DockerFabricClient, self).copy_resource(container, resource, local_filename)

    def cleanup_containers(self, include_initial=False, exclude=None, raise_on_error=False):
        """
        Identical to :meth:`dockermap.client.docker_util.DockerUtilityMixin.cleanup_containers` with additional logging.
        """
        self.push_log("Generating list of stopped containers.")
        super(DockerFabricClient, self).cleanup_containers(include_initial=include_initial, exclude=exclude,
                                                           raise_on_error=raise_on_error)

    def cleanup_images(self, remove_old=False, keep_tags=None, raise_on_error=False):
        """
        Identical to :meth:`dockermap.client.docker_util.DockerUtilityMixin.cleanup_images` with additional logging.
        """
        self.push_log("Checking images for dependent images and containers.")
        super(DockerFabricClient, self).cleanup_images(remove_old=remove_old, keep_tags=keep_tags,
                                                       raise_on_error=raise_on_error)

    def get_container_names(self):
        """
        Identical to :meth:`dockermap.client.docker_util.DockerUtilityMixin.get_container_names` with additional
        logging.
        """
        self.push_log("Fetching container list.")
        return super(DockerFabricClient, self).get_container_names()

    def get_image_tags(self):
        """
        Identical to :meth:`dockermap.client.docker_util.DockerUtilityMixin.get_image_tags` with additional logging.
        """
        self.push_log("Fetching image list.")
        return super(DockerFabricClient, self).get_image_tags()

    def import_image(self, image=None, tag='latest', **kwargs):
        """
        Identical to :meth:`docker.api.image.ImageApiMixin.import_image` with additional logging.
        """
        self.push_log("Fetching image '{0}' from registry.".format(image))
        return super(DockerFabricClient, self).import_image(image=image, tag=tag, **kwargs)

    def login(self, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.login` with two enhancements:

        * additional logging;
        * login parameters can be passed through ``kwargs``, or set as default using the following ``env``
          variables:

          * ``env.docker_registry_user`` (kwarg: ``username``),
          * ``env.docker_registry_password`` (kwarg: ``password``),
          * ``env.docker_registry_mail`` (kwarg: ``email``),
          * ``env.docker_registry_repository`` (kwarg: ``registry``),
          * ``env.docker_registry_insecure`` (kwarg: ``insecure_registry``).
        """
        c_user = kwargs.pop('username', env.get('docker_registry_user'))
        c_pass = kwargs.pop('password', env.get('docker_registry_password'))
        c_mail = kwargs.pop('email', env.get('docker_registry_mail'))
        c_registry = kwargs.pop('registry', env.get('docker_registry_repository'))
        c_insecure = kwargs.pop('insecure_registry', env.get('docker_registry_insecure'))
        if super(DockerFabricClient, self).login(c_user, password=c_pass, email=c_mail, registry=c_registry,
                                                 insecure_registry=c_insecure, **kwargs):
            self.push_log("Login at registry '{0}' succeeded.".format(c_registry))
            return True
        self.push_log("Login at registry '{0}' failed.".format(c_registry))
        return False

    def pull(self, repository, tag=None, stream=True, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.pull` with two enhancements:

        * additional logging;
        * the ``insecure_registry`` flag can be passed through ``kwargs``, or set as default using
          ``env.docker_registry_insecure``.
        """
        c_insecure = kwargs.pop('insecure_registry', env.get('docker_registry_insecure'))
        if 'raise_on_error' not in kwargs:
            kwargs['raise_on_error'] = True
        try:
            return super(DockerFabricClient, self).pull(repository, tag=tag, stream=stream,
                                                        insecure_registry=c_insecure, **kwargs)
        except DockerStatusError as e:
            error(e.message)

    def push(self, repository, stream=True, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.push` with two enhancements:

        * additional logging;
        * the ``insecure_registry`` flag can be passed through ``kwargs``, or set as default using
          ``env.docker_registry_insecure``.
        """
        c_insecure = kwargs.pop('insecure_registry', env.get('docker_registry_insecure'))
        kwargs['raise_on_error'] = True
        try:
            return super(DockerFabricClient, self).push(repository, stream=stream, insecure_registry=c_insecure,
                                                        **kwargs)
        except DockerStatusError as e:
            error(e.message)

    def restart(self, container, **kwargs):
        """
        Identical to :meth:`docker.api.container.ContainerApiMixin.restart` with additional logging.
        """
        self.push_log("Restarting container '{0}'.".format(container))
        super(DockerFabricClient, self).restart(container, **kwargs)

    def remove_all_containers(self, **kwargs):
        """
        Identical to :meth:`dockermap.client.docker_util.DockerUtilityMixin.remove_all_containers` with additional
        logging.
        """
        self.push_log("Fetching container list.")
        super(DockerFabricClient, self).remove_all_containers(**kwargs)

    def remove_container(self, container, raise_on_error=False, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.remove_container` with additional logging.
        """
        self.push_log("Removing container '{0}'.".format(container))
        super(DockerFabricClient, self).remove_container(container, raise_on_error=raise_on_error, **kwargs)

    def remove_image(self, image, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.remove_image` with additional logging.
        """
        self.push_log("Removing image '{0}'.".format(image))
        super(DockerFabricClient, self).remove_image(image, **kwargs)

    def save_image(self, image, local_filename):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.save_image` with additional logging.
        """
        self.push_log("Receiving tarball for image '{0}' and storing as '{1}'".format(image, local_filename))
        super(DockerFabricClient, self).save_image(image, local_filename)

    def start(self, container, **kwargs):
        """
        Identical to :meth:`docker.api.container.ContainerApiMixin.start` with additional logging.
        """
        self.push_log("Starting container '{0}'.".format(container))
        super(DockerFabricClient, self).start(container, **kwargs)

    def stop(self, container, **kwargs):
        """
        Identical to :meth:`dockermap.client.base.DockerClientWrapper.stop` with additional logging.
        """
        self.push_log("Stopping container '{0}'.".format(container))
        super(DockerFabricClient, self).stop(container, **kwargs)

    def wait(self, container, **kwargs):
        """
        Identical to :meth:`docker.api.container.ContainerApiMixin.wait` with additional logging.
        """
        self.push_log("Waiting for container '{0}'.".format(container))
        super(DockerFabricClient, self).wait(container, **kwargs)

    def run_cmd(self, command):
        sudo(command)


class DockerClientConfiguration(FabricClientConfiguration):
    init_kwargs = FabricClientConfiguration.init_kwargs + ('tunnel_remote_port', 'tunnel_local_port')
    client_constructor = DockerFabricClient


class DockerFabricApiConnections(DockerConnectionDict):
    configuration_class = DockerClientConfiguration


class ContainerApiFabricClient(FabricContainerClient):
    configuration_class = DockerClientConfiguration


# Still defined here for backwards compatibility.
docker_fabric = DockerFabricApiConnections().get_connection
container_fabric = ContainerApiFabricClient



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric.api import sudo
from fabric.utils import error

from dockermap.shortcuts import addgroup, adduser, assignuser
from .output import single_line_stdout, check_int


def get_group_id(groupname):
    """
    Returns the group id to a given group name. Returns ``None`` if the group does not exist.

    :param groupname: Group name.
    :type groupname: unicode
    :return: Group id.
    :rtype: int
    """
    gid = single_line_stdout('id -g {0}'.format(groupname), expected_errors=(1,), shell=False)
    return check_int(gid)


def get_user_id(username):
    """
    Returns the user id to a given user name. Returns ``None`` if the user does not exist.

    :param username: User name.
    :type username: unicode
    :return: User id.
    :rtype: int
    """
    uid = single_line_stdout('id -u {0}'.format(username), expected_errors=(1,), shell=False)
    return check_int(uid)


def get_user_groups(username):
    """
    Returns the list if group names for a given user name, omitting the default group.
    Returns ``None`` if the user does not exist.

    :param username: User name.
    :type username: unicode
    :return: Group names.
    :rtype: list
    """
    out = single_line_stdout('groups {0}'.format(username))
    if out:
        return out.split()[2:]
    return None


def create_group(groupname, gid, system=True):
    """
    Creates a new user group with a specific id.

    :param groupname: Group name.
    :type groupname: unicode
    :param gid: Group id.
    :type gid: int or unicode
    :param system: Creates a system group.
    """
    sudo(addgroup(groupname, gid, system))


def create_user(username, uid, system=False, no_login=True, no_password=False, group=False, gecos=None):
    """
    Creates a new user with a specific id.

    :param username: User name.
    :type username: unicode
    :param uid: User id.
    :type uid: int or unicode
    :param system: Creates a system user.
    :type system: bool
    :param no_login: Disallow login of this user and group, and skip creating the home directory. Default is ``True``.
    :type no_login: bool
    :param no_password: Do not set a password for the new user.
    :type: no_password: bool
    :param group: Create a group with the same id.
    :type group: bool
    :param gecos: Provide GECOS info and suppress prompt.
    :type gecos: unicode
    """
    sudo(adduser(username, uid, system, no_login, no_password, group, gecos))


def assign_user_groups(username, groupnames):
    """
    Assigns a user to a set of groups. User and group need to exists. The new groups are appended to existing group
    assignments.

    :param username: User name.
    :type username: unicode
    :param groupnames: Group names.
    :type groupnames: iterable
    """
    sudo(assignuser(username, groupnames))


def get_or_create_group(groupname, gid_preset, system=False, id_dependent=True):
    """
    Returns the id for the given group, and creates it first in case it does not exist.

    :param groupname: Group name.
    :type groupname: unicode
    :param gid_preset: Group id to set if a new group is created.
    :type gid_preset: int or unicode
    :param system: Create a system group.
    :type system: bool
    :param id_dependent: If the group exists, but its id does not match `gid_preset`, an error is thrown.
    :type id_dependent: bool
    :return: Group id of the existing or new group.
    :rtype: int
    """
    gid = get_group_id(groupname)
    if gid is None:
        create_group(groupname, gid_preset, system)
        return gid_preset
    elif id_dependent and gid != gid_preset:
        error("Present group id '{0}' does not match the required id of the environment '{1}'.".format(gid, gid_preset))
    return gid


def get_or_create_user(username, uid_preset, groupnames=[], system=False, no_password=False, no_login=True,
                       gecos=None, id_dependent=True):
    """
    Returns the id of the given user name, and creates it first in case it does not exist. A default group is created
    as well.

    :param username: User name.
    :type username: unicode
    :param uid_preset: User id to set in case a new user is created.
    :type uid_preset: int or unicode
    :param groupnames: Additional names of groups to assign the user to. If the user exists, these will be appended to
      existing group assignments.
    :type groupnames: iterable
    :param system: Create a system user.
    :type system: bool
    :param no_login: Disallow login of this user and group, and skip creating the home directory. Default is ``True``.
    :type no_login: bool
    :param no_password: Do not set a password for the new user.
    :type: no_password: bool
    :param gecos: Provide GECOS info and suppress prompt.
    :type gecos: unicode
    :param id_dependent: If the user exists, but its id does not match `uid_preset`, an error is thrown.
    :type id_dependent: bool
    :return:
    """
    uid = get_user_id(username)
    gid = get_group_id(username)
    if id_dependent and gid is not None and gid != uid_preset:
        error("Present group id '{0}' does not match the required id of the environment '{1}'.".format(gid, uid_preset))
    if gid is None:
        create_group(username, uid_preset, system)
        gid = uid_preset
    if uid is None:
        create_user(username, gid, system, no_login, no_password, False, gecos)
        if groupnames:
            assign_user_groups(username, groupnames)
        return uid
    elif id_dependent and uid != uid_preset:
        error("Present user id '{0}' does not match the required id of the environment '{1}'.".format(uid, uid_preset))
    current_groups = get_user_groups(username)
    new_groups = set(groupnames).discard(tuple(current_groups))
    if new_groups:
        assign_user_groups(username, new_groups)
    return uid



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import posixpath

from fabric.api import get, put, puts, task
from fabric.utils import error
import six

from .apiclient import container_fabric
from .utils.files import temp_dir


@task
def perform(action_name, map_name, config_name, **kwargs):
    """
    Performs an action on the given container map and configuration.

    :param action_name: Name of the action (e.g. ``update``).
    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param kwargs: Keyword arguments for the action implementation.
    """
    cf = container_fabric()
    cf.call(action_name, config_name, map_name=map_name, **kwargs)


@task
def create(map_name, config_name, instances=None, **kwargs):
    """
    Creates a container and its dependencies.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().create(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def start(map_name, config_name, instances=None, **kwargs):
    """
    Starts a container and its dependencies.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().start(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def stop(map_name, config_name, instances=None, **kwargs):
    """
    Stops a container and its dependents.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().stop(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def remove(map_name, config_name, instances=None, **kwargs):
    """
    Removes a container and its dependents.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().remove(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def restart(map_name, config_name, instances=None, **kwargs):
    """
    Restarts a container and starts its dependencies if necessary.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().restart(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def startup(map_name, config_name, instances=None, **kwargs):
    """
    Creates and starts a container and its dependencies.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().startup(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def shutdown(map_name, config_name, instances=None, **kwargs):
    """
    Stops and removes a container and its dependents.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().shutdown(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def update(map_name, config_name, instances=None, **kwargs):
    """
    Updates a container and its dependencies. Creates and starts containers as necessary.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param instances: Optional instance names.
    :param kwargs: Keyword arguments to the action implementation.
    """
    container_fabric().update(config_name, instances=instances, map_name=map_name, **kwargs)


@task
def script(map_name, config_name, script_path, fail_nonzero=False, upload_dir=False, **kwargs):
    """
    Runs a script inside a container, which is created with all its dependencies. The container is removed after it
    has been run, whereas the dependencies are not destroyed. The output is printed to the console.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param script_path: Local path to the script file.
    :param fail_nonzero: Fail if the script returns with a nonzero exit code.
    :param upload_dir: Upload the entire parent directory of the script file to the remote.
    :param kwargs: Additional keyword arguments to the run_script action.
    """
    full_script_path = os.path.abspath(script_path)
    prefix, name = os.path.split(full_script_path)
    with temp_dir() as remote_tmp:
        if upload_dir:
            prefix_path, prefix_name = os.path.split(prefix)
            remote_script = posixpath.join(remote_tmp, prefix_name, name)
            put(prefix, remote_tmp, mirror_local_mode=True)
        else:
            remote_script = posixpath.join(remote_tmp, name)
            put(script_path, remote_script, mirror_local_mode=True)
        results = container_fabric().run_script(config_name, map_name=map_name, script_path=remote_script, **kwargs)
        for client, res in six.iteritems(results):
            puts("Exit code: {0}".format(res['exit_code']))
            if res['exit_code'] == 0 or not fail_nonzero:
                puts(res['log'])
            else:
                error(res['log'])


@task
def single_cmd(map_name, config_name, command, fail_nonzero=False, download_result=None, **kwargs):
    """
    Runs a script inside a container, which is created with all its dependencies. The container is removed after it
    has been run, whereas the dependencies are not destroyed. The output is printed to the console.

    :param map_name: Container map name.
    :param config_name: Container configuration name.
    :param fail_nonzero: Fail if the script returns with a nonzero exit code.
    :param download_result: Download any results that the command has written back to a temporary directory.
    :param kwargs: Additional keyword arguments to the run_script action.
    """
    with temp_dir() as remote_tmp:
        if 'command_format' not in kwargs:
            kwargs['command_format'] = ['-c', command]
        results = container_fabric().run_script(config_name, map_name=map_name, script_path=remote_tmp, **kwargs)
        for client, res in six.iteritems(results):
            puts("Exit code: {0}".format(res['exit_code']))
            if res['exit_code'] == 0 or not fail_nonzero:
                puts(res['log'])
            else:
                error(res['log'])
            if download_result:
                get(posixpath.join(remote_tmp, '*'), local_path=download_result)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric.context_managers import documented_contextmanager
from dockerfabric.apiclient import docker_fabric


@documented_contextmanager
def temp_container(image, no_op_cmd='/bin/true', create_kwargs=None, start_kwargs=None):
    """
    Creates a temporary container, which can be used e.g. for copying resources. The container is removed once it
    is no longer needed. Note that ``no_op_cmd`` needs to be set appropriately, since the method will wait for the
    container to finish before copying resources.

    :param image: Image name or id to create the container from.
    :type image: unicode
    :param no_op_cmd: Dummy-command to run, only for being able to access the container.
    :type no_op_cmd: unicode
    :param create_kwargs: Additional kwargs for creating the container. The ``entrypoint`` will be set to ``no_op_cmd``.
    :type create_kwargs: dict
    :param start_kwargs: Additional kwargs for starting the container. ``restart_policy`` will be set to ``None``.
    :type start_kwargs: dict
    :return: Id of the temporary container.
    :rtype: unicode
    """
    df = docker_fabric()
    create_kwargs = create_kwargs.copy() if create_kwargs else dict()
    start_kwargs = start_kwargs.copy() if start_kwargs else dict()
    create_kwargs.update(entrypoint=no_op_cmd)
    start_kwargs.update(restart_policy=None)
    container = df.create_container(image, **create_kwargs)['Id']
    df.start(container, **start_kwargs)
    df.wait(container)
    yield container
    df.remove_container(container)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from datetime import datetime
import itertools
import os
from fabric.api import cd, env, get, put, run, runs_once, sudo, task
from fabric.utils import error, puts, fastprint
import six

from dockermap.shortcuts import curl, untargz
from dockermap.utils import expand_path
from . import DEFAULT_SOCAT_VERSION, cli
from .api import docker_fabric
from .utils.files import temp_dir
from .utils.net import get_ip4_address, get_ip6_address
from .utils.output import stdout_result
from .utils.users import assign_user_groups


SOCAT_URL = 'http://www.dest-unreach.org/socat/download/socat-{0}.tar.gz'
IMAGE_COLUMNS = ('Id', 'RepoTags', 'ParentId', 'Created', 'VirtualSize', 'Size')
CONTAINER_COLUMNS = ('Id', 'Names', 'Image', 'Command', 'Ports', 'Status', 'Created')


def _format_output_table(data_dict, columns, full_ids=False, full_cmd=False, short_image=False):
    def _format_port(port_dict):
        if 'PublicPort' in port_dict and 'IP' in port_dict:
            return '{IP}:{PublicPort}->{PrivatePort}/{Type}'.format(**port_dict)
        return '{PrivatePort}/{Type}'.format(**port_dict)

    def _get_column(item, column):
        data = item.get(column, '')
        if isinstance(data, list):
            if column == 'Ports':
                return map(_format_port, data)
            return data
        if column in ('Id', 'ParentId') and not full_ids:
            return data[:12],
        if column == 'Created':
            return datetime.utcfromtimestamp(data).isoformat(),
        if column == 'Command' and not full_cmd:
            return data[:25],
        if column == 'Image' and short_image:
            __, __, i_name = data.rpartition('/')
            return i_name,
        return unicode(data),

    def _max_len(col_data):
        if col_data:
            return max(map(len, col_data))
        return 0

    puts('')
    rows = [[[c] for c in columns]]
    rows.extend([_get_column(i, col) for col in columns] for i in data_dict)
    col_lens = map(max, (map(_max_len, c) for c in zip(*rows)))
    row_format = '  '.join('{{{0}:{1}}}'.format(i, l) for i, l in enumerate(col_lens))
    for row in rows:
        for c in itertools.izip_longest(*row, fillvalue=''):
            fastprint(row_format.format(*c), end='\n', flush=False)
    fastprint('', flush=True)


@task
def install_docker_ubuntu(skip_group_assignment=False):
    """
    Installs Docker on a remote machine running Ubuntu and adds the current user to the ``docker`` user group.

    :param skip_group_assignment: If set to ``True``, skips the assignment to the ``docker`` group.
    :type skip_group_assignment: bool
    """
    sudo('apt-key adv --keyserver hkp://p80.pool.sks-keyservers.net:80 '
         '--recv-keys 58118E89F3A912897C070ADBF76221572C52609D')
    sudo('echo deb https://apt.dockerproject.org/repo ubuntu-`lsb_release -c -s` main > '
         '/etc/apt/sources.list.d/docker.list')
    sudo('apt-get update -o Dir::Etc::sourcelist="sources.list.d/docker.list" -o Dir::Etc::sourceparts="-" '
         '-o APT::Get::List-Cleanup="0"')
    sudo('apt-get -y install docker-engine')
    if not skip_group_assignment:
        assign_user_groups(env.user, ['docker'])


@task
def install_docker_centos(skip_group_assignment=False):
    """
    Installs Docker on a remote machine running CentOS and adds the current user to the ``docker`` user group.

    :param skip_group_assignment: If set to ``True``, skips the assignment to the ``docker`` group.
    :type skip_group_assignment: bool
    """
    sudo("tee /etc/yum.repos.d/docker.repo <<-'EOF'\n"
         "[dockerrepo]\n"
         "name=Docker Repository\n"
         "baseurl=https://yum.dockerproject.org/repo/main/centos/$releasever/\n"
         "enabled=1\n"
         "gpgcheck=1\n"
         "gpgkey=https://yum.dockerproject.org/gpg\n"
         "EOF\n")
    sudo('yum install -y docker-engine')
    if not skip_group_assignment:
        assign_user_groups(env.user, ['docker'])


def _build_socat():
    with temp_dir() as remote_tmp:
        socat_version = env.get('socat_version', DEFAULT_SOCAT_VERSION)
        src_dir = '{0}/socat-{1}'.format(remote_tmp, socat_version)
        src_file = '.'.join((src_dir, 'tar.gz'))
        run(curl(SOCAT_URL.format(socat_version), src_file))
        run(untargz(src_file, remote_tmp))
        with cd(src_dir):
            run('./configure')
            run('make')
            sudo('make install')


@task
def build_socat_ubuntu():
    """
    Downloads and installs the tool `socat` from source on Ubuntu.
    """
    sudo('apt-get update')
    sudo('apt-get -y install gcc make')
    _build_socat()


@task
def build_socat_centos():
    """
    Downloads and installs the tool `socat` from source on CentOS.
    """
    sudo('yum install -y gcc make')
    _build_socat()


@task
@runs_once
def fetch_socat(local):
    """
    Fetches the `socat` binary from a remote host.

    :param local: Local path to copy the file to, or local file path.
    :type local: unicode
    """
    remote_file = '/usr/local/bin/socat'
    local_file = expand_path(local)
    if os.path.exists(local_file) and not os.path.isfile(local_file):
        local_file = os.path.join(local, 'socat')
    get(remote_file, local_file)


@task
def install_socat(src):
    """
    Places the `socat` binary on a remote host.

    :param src: Local directory that contains the source file, or path to the file itself.
    :type src: unicode
    """
    src_file = expand_path(src)
    if os.path.exists(src_file) and not os.path.isfile(src_file):
        src_file = os.path.join(src_file, 'socat')
        if not os.path.exists(src_file):
            error("Socat cannot be found in the provided path ({0} or {1}).".format(src, src_file))
    dest_file = '/usr/local/bin/socat'
    put(src_file, dest_file, use_sudo=True, mode='0755')


@task
def reset_socat(use_sudo=False):
    """
    Finds and closes all processes of `socat`.

    :param use_sudo: Use `sudo` command. As Docker-Fabric does not run `socat` with `sudo`, this is by default set to
      ``False``. Setting it to ``True`` could unintentionally remove instances from other users.
    :type use_sudo: bool
    """
    output = stdout_result('ps -o pid -C socat', quiet=True)
    pids = output.split('\n')[1:]
    puts("Removing process(es) with id(s) {0}.".format(', '.join(pids)))
    which = sudo if use_sudo else run
    which('kill {0}'.format(' '.join(pids)), quiet=True)


@task
def version():
    """
    Shows version information of the remote Docker service, similar to ``docker version``.
    """
    output = docker_fabric().version()
    col_len = max(map(len, output.keys())) + 1
    puts('')
    for k, v in six.iteritems(output):
        fastprint('{0:{1}} {2}'.format(''.join((k, ':')), col_len, v), end='\n', flush=False)
    fastprint('', flush=True)


@task
def get_ip(interface_name='docker0'):
    """
    Shows the IP4 address of a network interface.

    :param interface_name: Name of the network interface. Default is ``docker0``.
    :type interface_name: unicode
    """
    puts(get_ip4_address(interface_name))


@task
def get_ipv6(interface_name='docker0', expand=False):
    """
    Shows the IP6 address of a network interface.

    :param interface_name: Name of the network interface. Default is ``docker0``.
    :type interface_name: unicode
    :param expand: Expand the abbreviated IP6 address. Default is ``False``.
    :type expand: bool
    """
    puts(get_ip6_address(interface_name, expand=expand))


@task
def list_images(list_all=False, full_ids=False):
    """
    Lists images on the Docker remote host, similar to ``docker images``.

    :param list_all: Lists all images (e.g. dependencies). Default is ``False``, only shows named images.
    :type list_all: bool
    :param full_ids: Shows the full ids. When ``False`` (default) only shows the first 12 characters.
    :type full_ids: bool
    """
    images = docker_fabric().images(all=list_all)
    _format_output_table(images, IMAGE_COLUMNS, full_ids)


@task
def list_containers(list_all=True, short_image=True, full_ids=False, full_cmd=False):
    """
    Lists containers on the Docker remote host, similar to ``docker ps``.

    :param list_all: Shows all containers. Default is ``False``, which omits exited containers.
    :type list_all: bool
    :param short_image: Hides the repository prefix for preserving space. Default is ``True``.
    :type short_image: bool
    :param full_ids: Shows the full image ids. When ``False`` (default) only shows the first 12 characters.
    :type full_ids: bool
    :param full_cmd: Shows the full container command. When ``False`` (default) only shows the first 25 characters.
    :type full_cmd: bool
    """
    containers = docker_fabric().containers(all=list_all)
    _format_output_table(containers, CONTAINER_COLUMNS, full_ids, full_cmd, short_image)


@task
def cleanup_containers(include_initial=False):
    """
    Removes all containers that have finished running.
    """
    docker_fabric().cleanup_containers(include_initial=include_initial)


@task
def cleanup_images(remove_old=False):
    """
    Removes all images that have no name, and that are not references as dependency by any other named image.

    :param remove_old: Also remove images that do have a name, but no `latest` tag.
    :type remove_old: bool
    """
    keep_tags = env.get('docker_keep_tags')
    docker_fabric().cleanup_images(remove_old=remove_old, keep_tags=keep_tags)


@task
def remove_all_containers():
    """
    Stops and removes all containers from the remote. Use with caution outside of a development environment!
    :return:
    """
    docker_fabric().remove_all_containers()


@task
@runs_once
def save_image(image, filename=None):
    """
    Saves a Docker image from the remote to a local files. For performance reasons, uses the Docker command line client
    on the host, generates a gzip-tarball and downloads that.

    :param image: Image name or id.
    :type image: unicode
    :param filename: File name to store the local file. If not provided, will use ``<image>.tar.gz`` in the current
      working directory.
    :type filename: unicode
    """
    local_name = filename or '{0}.tar.gz'.format(image)
    cli.save_image(image, local_name)


@task
def load_image(filename, timeout=120):
    """
    Uploads an image from a local file to a Docker remote. Note that this temporarily has to extend the service timeout
    period.

    :param filename: Local file name.
    :type filename: unicode
    :param timeout: Timeout in seconds to set temporarily for the upload.
    :type timeout: int
    """
    c = docker_fabric()
    with open(expand_path(filename), 'r') as f:
        _timeout = c._timeout
        c._timeout = timeout
        try:
            c.load_image(f)
        finally:
            c._timeout = _timeout



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import posixpath

from fabric.api import cd, env, fastprint, get, put, run, settings, sudo
from fabric.network import needs_host

from dockermap.api import USE_HC_MERGE
from dockermap.client.cli import (DockerCommandLineOutput, parse_containers_output, parse_inspect_output,
                                  parse_images_output, parse_version_output)
from dockermap.client.docker_util import DockerUtilityMixin
from dockermap.shortcuts import chmod, chown, targz, mkdir

from .base import DockerConnectionDict, FabricContainerClient, FabricClientConfiguration
from .utils.containers import temp_container
from .utils.files import temp_dir, is_directory


class DockerCliClient(DockerUtilityMixin):
    """
    Docker client for Fabric using the command line interface on a remote host.

    :param cmd_prefix: Custom prefix to prepend to the Docker command line.
    :type cmd_prefix: unicode
    :param default_bin: Docker binary to use. If not set, uses ``docker``.
    :type default_bin: unicode
    :param base_url: URL to connect to; if not set, will refer to ``env.docker_base_url`` or use ``None``, which by
     default attempts a connection on a Unix socket at ``/var/run/docker.sock``.
    :type base_url: unicode
    :param tls: Whether to use TLS on the connection to the Docker service.
    :type tls: bool
    :param use_sudo: Whether to use ``sudo`` when performing Docker commands.
    :type use_sudo: bool
    :param debug: If set to ``True``, echoes each command and its console output. Some commands are echoed either way
     for some feedback.
    :type debug: bool
    """
    def __init__(self, cmd_prefix=None, default_bin=None, base_url=None, tls=None, use_sudo=None, debug=None):
        super(DockerCliClient, self).__init__()
        base_url = base_url or env.get('docker_base_url')
        if base_url:
            cmd_args = ['-H {0}'.format(base_url)]
        else:
            cmd_args = []
        if tls or (tls is None and env.get('docker_tls')):
            cmd_args.append('--tls')
        self._out = DockerCommandLineOutput(cmd_prefix or env.get('docker_cli_prefix'),
                                            default_bin or env.get('docker_cli_bin', 'docker'), cmd_args or None)
        if use_sudo or (use_sudo is None and env.get('docker_cli_sudo')):
            self._call_method = sudo
        else:
            self._call_method = run
        self._quiet = not (debug or (debug is None and env.get('docker_cli_debug')))
        self.api_version = None
        self._update_api_version()

    def _call(self, cmd, quiet=False):
        if cmd:
            return self._call_method(cmd, shell=False, quiet=quiet and self._quiet)
        return None

    def create_container(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('create_container', *args, **kwargs)
        return {'Id': self._call(cmd_str)}

    def start(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('start', *args, **kwargs)
        self._call(cmd_str)

    def restart(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('restart', *args, **kwargs)
        self._call(cmd_str)

    def stop(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('stop', *args, **kwargs)
        self._call(cmd_str)

    def remove_container(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('remove_container', *args, **kwargs)
        self._call(cmd_str)

    def remove_image(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('remove_image', *args, **kwargs)
        self._call(cmd_str)

    def kill(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('kill', *args, **kwargs)
        self._call(cmd_str)

    def wait(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('wait', *args, **kwargs)
        self._call(cmd_str)

    def containers(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('containers', *args, **kwargs)
        res = self._call(cmd_str, quiet=True)
        return parse_containers_output(res)

    def inspect_container(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('inspect_container', *args, **kwargs)
        res = self._call(cmd_str, quiet=True)
        return parse_inspect_output(res)

    def images(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('images', *args, **kwargs)
        res = self._call(cmd_str, quiet=True)
        return parse_images_output(res)

    def pull(self, repository, tag=None, **kwargs):
        repo_tag = '{0}:{1}'.format(repository, tag) if tag else repository
        cmd_str = self._out.get_cmd('pull', repo_tag, **kwargs)
        self._call(cmd_str)

    def push(self, repository, tag=None, **kwargs):
        repo_tag = '{0}:{1}'.format(repository, tag) if tag else repository
        cmd_str = self._out.get_cmd('push', repo_tag, **kwargs)
        self._call(cmd_str)

    def exec_create(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('exec_create', *args, **kwargs)
        self._call(cmd_str)

    def exec_start(self, *args, **kwargs):
        cmd_str = self._out.get_cmd('exec_start', *args, **kwargs)
        self._call(cmd_str)

    def tag(self, image, repository, tag=None, **kwargs):
        if tag:
            repo_tag = '{0}:{1}'.format(repository, tag)
        else:
            repo_tag = repository
        cmd_str = self._out.get_cmd('tag', image, repo_tag, **kwargs)
        return self._call(cmd_str)

    def logs(self, *args, **kwargs):
        kwargs.pop('stream', None)
        cmd_str = self._out.get_cmd('logs', *args, **kwargs)
        return self._call(cmd_str, quiet=True)

    def login(self, **kwargs):
        for key, variable in [
            ('username', 'user'),
            ('password', 'password'),
            ('email', 'mail'),
            ('registry', 'repository'),
            ('insecure_registry', 'insecure')
        ]:
            if key not in kwargs:
                env_value = env.get('docker_registry_{0}'.format(variable))
                if env_value:
                    kwargs[key] = env_value
        registry = kwargs.pop('registry', env.get('docker_registry_repository'))
        if registry:
            cmd_str = self._out.get_cmd('login', registry, **kwargs)
        else:
            cmd_str = self._out.get_cmd('login', **kwargs)
        res = self._call(cmd_str, quiet=True)
        lines = res.splitlines()
        fastprint(lines)
        return 'Login Succeeded' in lines

    def build(self, tag, add_latest_tag=False, add_tags=None, raise_on_error=False, **kwargs):
        try:
            context = kwargs.pop('fileobj')
        except KeyError:
            raise ValueError("'fileobj' needs to be provided. Using 'path' is currently not implemented.")
        for a in ['custom_context', 'encoding']:
            kwargs.pop(a, None)

        with temp_dir() as remote_tmp:
            remote_fn = posixpath.join(remote_tmp, 'context')
            put(context, remote_fn)
            cmd_str = self._out.get_cmd('build', '- <', remote_fn, tag=tag, **kwargs)
            with settings(warn_only=not raise_on_error):
                res = self._call(cmd_str)
        if res:
            last_log = res.splitlines()[-1]
            if last_log and last_log.startswith('Successfully built '):
                image_id = last_log[19:]  # Remove prefix
                self.add_extra_tags(image_id, tag, add_tags, add_latest_tag)
                return image_id
        return None

    def version(self, **kwargs):
        kwargs.pop('api_version', None)
        cmd_str = self._out.get_cmd('version')
        res = self._call(cmd_str, quiet=True)
        version_dict = parse_version_output(res)
        return version_dict

    def push_log(self, info, level, *args, **kwargs):
        pass

    def _update_api_version(self):
        if self.api_version and self.api_version != 'auto':
            return
        version_dict = self.version()
        if 'APIVersion' in version_dict:
            self.api_version = version_dict['APIVersion']
        elif 'ApiVersion' in version_dict:
            self.api_version = version_dict['ApiVersion']


class DockerCliConfig(FabricClientConfiguration):
    init_kwargs = 'base_url', 'tls', 'cmd_prefix', 'default_bin', 'use_sudo', 'debug'
    client_constructor = DockerCliClient

    def update_settings(self, **kwargs):
        super(DockerCliConfig, self).update_settings(**kwargs)
        self.use_host_config = USE_HC_MERGE


class DockerCliConnections(DockerConnectionDict):
    configuration_class = DockerCliConfig


class ContainerCliFabricClient(FabricContainerClient):
    configuration_class = DockerCliConfig


docker_cli = DockerCliConnections().get_connection
container_cli = ContainerCliFabricClient


@needs_host
def copy_resource(container, resource, local_filename, contents_only=True):
    """
    Copies a resource from a container to a compressed tarball and downloads it.

    :param container: Container name or id.
    :type container: unicode
    :param resource: Name of resource to copy.
    :type resource: unicode
    :param local_filename: Path to store the tarball locally.
    :type local_filename: unicode
    :param contents_only: In case ``resource`` is a directory, put all contents at the root of the tar file. If this is
      set to ``False``, the directory itself will be at the root instead.
    :type contents_only: bool
    """
    with temp_dir() as remote_tmp:
        base_name = os.path.basename(resource)
        copy_path = posixpath.join(remote_tmp, 'copy_tmp')
        run(mkdir(copy_path, check_if_exists=True))
        remote_name = posixpath.join(copy_path, base_name)
        archive_name = 'container_{0}.tar.gz'.format(container)
        archive_path = posixpath.join(remote_tmp, archive_name)
        run('docker cp {0}:{1} {2}'.format(container, resource, copy_path), shell=False)
        if contents_only and is_directory(remote_name):
            src_dir = remote_name
            src_files = '*'
        else:
            src_dir = copy_path
            src_files = base_name
        with cd(src_dir):
            run(targz(archive_path, src_files))
        get(archive_path, local_filename)


@needs_host
def copy_resources(src_container, src_resources, storage_dir, dst_directories=None, apply_chown=None, apply_chmod=None):
    """
    Copies files and directories from a Docker container. Multiple resources can be copied and additional options are
    available than in :func:`copy_resource`. Unlike in :func:`copy_resource`, Resources are copied as they are and not
    compressed to a tarball, and they are left on the remote machine.

    :param src_container: Container name or id.
    :type src_container: unicode
    :param src_resources: Resources, as (file or directory) names to copy.
    :type src_resources: iterable
    :param storage_dir: Remote directory to store the copied objects in.
    :type storage_dir: unicode
    :param dst_directories: Optional dictionary of destination directories, in the format ``resource: destination``. If
      not set, resources will be in the same relative structure to one another as inside the container. For setting a
      common default, use ``*`` as the resource key.
    :type dst_directories: dict
    :param apply_chown: Owner to set for the copied resources. Can be a user name or id, group name or id, both in the
      notation ``user:group``, or as a tuple ``(user, group)``.
    :type apply_chown: unicode or tuple
    :param apply_chmod: File system permissions to set for the copied resources. Can be any notation as accepted by
      `chmod`.
    :type apply_chmod: unicode
    """
    def _copy_resource(resource):
        default_dest_path = generic_path if generic_path is not None else resource
        dest_path = directories.get(resource, default_dest_path).strip(posixpath.sep)
        head, tail = posixpath.split(dest_path)
        rel_path = posixpath.join(storage_dir, head)
        run(mkdir(rel_path, check_if_exists=True))
        run('docker cp {0}:{1} {2}'.format(src_container, resource, rel_path), shell=False)

    directories = dst_directories or {}
    generic_path = directories.get('*')
    for res in src_resources:
        _copy_resource(res)
    if apply_chmod:
        run(chmod(apply_chmod, storage_dir))
    if apply_chown:
        sudo(chown(apply_chown, storage_dir))


@needs_host
def isolate_and_get(src_container, src_resources, local_dst_dir, **kwargs):
    """
    Uses :func:`copy_resources` to copy resources from a container, but afterwards generates a compressed tarball
    and downloads it.

    :param src_container: Container name or id.
    :type src_container: unicode
    :param src_resources: Resources, as (file or directory) names to copy.
    :type src_resources: iterable
    :param local_dst_dir: Local directory to store the compressed tarball in. Can also be a file name; the default file
      name is ``container_<container name>.tar.gz``.
    :type local_dst_dir: unicode
    :param kwargs: Additional kwargs for :func:`copy_resources`.
    """
    with temp_dir() as remote_tmp:
        copy_path = posixpath.join(remote_tmp, 'copy_tmp')
        archive_path = posixpath.join(remote_tmp, 'container_{0}.tar.gz'.format(src_container))
        copy_resources(src_container, src_resources, copy_path, **kwargs)
        with cd(copy_path):
            sudo(targz(archive_path, '*'))
        get(archive_path, local_dst_dir)


@needs_host
def isolate_to_image(src_container, src_resources, dst_image, **kwargs):
    """
    Uses :func:`copy_resources` to copy resources from a container, but afterwards imports the contents into a new
    (otherwise empty) Docker image.

    :param src_container: Container name or id.
    :type src_container: unicode
    :param src_resources: Resources, as (file or directory) names to copy.
    :type src_resources: iterable
    :param dst_image: Tag for the new image.
    :type dst_image: unicode
    :param kwargs: Additional kwargs for :func:`copy_resources`.
    """
    with temp_dir() as remote_tmp:
        copy_resources(src_container, src_resources, remote_tmp, **kwargs)
        with cd(remote_tmp):
            sudo('tar -cz * | docker import - {0}'.format(dst_image))


@needs_host
def save_image(image, local_filename):
    """
    Saves a Docker image as a compressed tarball. This command line client method is a suitable alternative, if the
    Remove API method is too slow.

    :param image: Image id or tag.
    :type image: unicode
    :param local_filename: Local file name to store the image into. If this is a directory, the image will be stored
      there as a file named ``image_<Image name>.tar.gz``.
    """
    r_name, __, i_name = image.rpartition('/')
    i_name, __, __ = i_name.partition(':')
    with temp_dir() as remote_tmp:
        archive = posixpath.join(remote_tmp, 'image_{0}.tar.gz'.format(i_name))
        run('docker save {0} | gzip --stdout > {1}'.format(image, archive), shell=False)
        get(archive, local_filename)


@needs_host
def flatten_image(image, dest_image=None, no_op_cmd='/bin/true', create_kwargs={}, start_kwargs={}):
    """
    Exports a Docker image's file system and re-imports it into a new (otherwise new) image. Note that this does not
    transfer the image configuration. In order to gain access to the container contents, the image is started with a
    non-operational command, such as ``/bin/true``. The container is removed once the new image has been created.

    :param image: Image id or tag.
    :type image: unicode
    :param dest_image: Tag for the new image.
    :type dest_image: unicode
    :param no_op_cmd: Dummy command for starting temporary container.
    :type no_op_cmd: unicode
    :param create_kwargs: Optional additional kwargs for creating the temporary container.
    :type create_kwargs: dict
    :param start_kwargs: Optional additional kwargs for starting the temporary container.
    :type start_kwargs: dict
    """
    dest_image = dest_image or image
    with temp_container(image, no_op_cmd=no_op_cmd, create_kwargs=create_kwargs, start_kwargs=start_kwargs) as c:
        run('docker export {0} | docker import - {1}'.format(c, dest_image), shell=False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import select
import socket
import errno

from fabric.network import needs_host
from fabric.state import connections, env
from fabric.thread_handling import ThreadHandler

from .base import ConnectionDict, get_local_port


class LocalTunnels(ConnectionDict):
    """
    Cache for local tunnels to the remote machine.
    """
    def __getitem__(self, item):
        """
        :param item: Tuple of remote host, remote port, local port number, and local bind address.
        :type item: tuple
        :return: Local tunnel
        :rtype: LocalTunnel
        """
        def _connect_local_tunnel():
            bind_port = get_local_port(init_bind_port)
            tun = LocalTunnel(remote_port, remote_host, bind_port, bind_host)
            tun.connect()
            return tun

        remote_host, remote_port, bind_host, init_bind_port = item
        key = remote_host, remote_port
        return self.get_or_create_connection(key, _connect_local_tunnel)


local_tunnels = LocalTunnels()


def _forwarder(chan, sock):
    # Bidirectionally forward data between a socket and a Paramiko channel.
    try:
        while True:
            r, w, x = select.select([sock, chan], [], [], 1)
            if sock in r:
                data = sock.recv(1024)
                if len(data) == 0:
                    break
                chan.sendall(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                sock.sendall(data)
    except socket.error as e:
        #Sockets return bad file descriptor if closed.
        #Maybe there is a cleaner way of doing this?
        if e.errno not in (socket.EBADF, errno.ECONNRESET):
            raise
    except select.error as e:
        if e[0] != socket.EBADF:
            raise

    try:
        chan.close()
    except socket.error:
        pass

    try:
        sock.close()
    except socket.error:
        pass


class LocalTunnel(object):
    """
    Adapted from PR #939 of Fabric: https://github.com/fabric/fabric/pull/939

    Forward a local port to a given host and port on the remote side.

    :param remote_port: Remote port to forward connections to.
    :type remote_port: int
    :param remote_host: Host to connect to. Optional, default is ``localhost``.
    :type remote_host: unicode
    :param bind_port: Local port to bind to. Optional, default is same as ``remote_port``.
    :type bind_port: int
    :param bind_host: Local address to bind to. Optional, default is ``localhost``.
    """
    def __init__(self, remote_port, remote_host=None, bind_port=None, bind_host=None, remote_cmd=None):
        self.remote_port = remote_port
        self.remote_host = remote_host or 'localhost'
        self.bind_port = bind_port or remote_port
        self.bind_host = bind_host or 'localhost'
        self.remote_cmd = remote_cmd
        self.sockets = []
        self.channels = []
        self.threads = []
        self.listening_socket = None
        self.listening_thread = None

    def get_channel(self, transport, remote_addr, local_peer):
        channel = transport.open_channel('direct-tcpip', remote_addr, local_peer)
        if channel is None:
            raise Exception('Incoming request to %s:%d was rejected by the SSH server.' % remote_addr)
        return channel

    @needs_host
    def connect(self):
        def listener_thread_main(thead_sock, callback, *a, **kw):
            try:
                while True:
                    selsockets = select.select([thead_sock], [], [], 1)
                    if thead_sock in selsockets[0]:
                        callback(thead_sock, *a, **kw)
            except socket.error as e:
                #Sockets return bad file descriptor if closed.
                #Maybe there is a cleaner way of doing this?
                if e.errno != socket.EBADF:
                    raise
            except select.error as e:
                if e[0] != socket.EBADF:
                    raise

        listening_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        listening_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        listening_socket.bind((self.bind_host, self.bind_port))
        listening_socket.listen(1)

        def accept(listen_sock, transport, remote_addr):
            accept_sock, local_peer = listen_sock.accept()
            channel = self.get_channel(transport, remote_addr, local_peer)

            handler = ThreadHandler('fwd', _forwarder, channel, accept_sock)

            self.sockets.append(accept_sock)
            self.channels.append(channel)
            self.threads.append(handler)

        self.sockets = []
        self.channels = []
        self.threads = []
        self.listening_socket = listening_socket
        self.listening_thread = ThreadHandler('local_bind', listener_thread_main,
                                              listening_socket, accept,
                                              connections[env.host_string].get_transport(),
                                              (self.remote_host, self.remote_port))

    def close(self):
        for sock, chan, th in zip(self.sockets, self.channels, self.threads):
            sock.close()
            if not chan.closed:
                chan.close()
            th.thread.join()
            th.raise_if_needed()

        self.listening_socket.close()
        self.listening_thread.thread.join()
        self.listening_thread.raise_if_needed()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import shutil
import tarfile
import tempfile

from fabric.api import run, sudo
from fabric.context_managers import documented_contextmanager

from dockermap.shortcuts import rm, chmod, chown
from .output import single_line_stdout


def _safe_name(tarinfo):
    return tarinfo.name[0] != '/' and '..' not in tarinfo.name


def get_remote_temp():
    """
    Creates a temporary directory on the remote end. Uses the command ``mktemp`` to do so.

    :return: Path to the temporary directory.
    :rtype: unicode
    """
    return single_line_stdout('mktemp -d')


def remove_ignore(path, use_sudo=False, force=False):
    """
    Recursively removes a file or directory, ignoring any errors that may occur. Should only be used for temporary
    files that can be assumed to be cleaned up at a later point.

    :param path: Path to file or directory to remove.
    :type path: unicode
    :param use_sudo: Use the `sudo` command.
    :type use_sudo: bool
    :param force: Force the removal.
    :type force: bool
    """
    which = sudo if use_sudo else run
    which(rm(path, recursive=True, force=force), warn_only=True)


def is_directory(path, use_sudo=False):
    """
    Check if the remote path exists and is a directory.

    :param path: Remote path to check.
    :type path: unicode
    :param use_sudo: Use the `sudo` command.
    :type use_sudo: bool
    :return: `True` if the path exists and is a directory; `False` if it exists, but is a file; `None` if it does not
      exist.
    :rtype: bool or ``None``
    """
    result = single_line_stdout('if [[ -f {0} ]]; then echo 0; elif [[ -d {0} ]]; then echo 1; else echo -1; fi'.format(path), sudo=use_sudo, quiet=True)
    if result == '0':
        return False
    elif result == '1':
        return True
    else:
        return None


@documented_contextmanager
def temp_dir(apply_chown=None, apply_chmod=None, remove_using_sudo=None, remove_force=False):
    """
    Creates a temporary directory on the remote machine. The directory is removed when no longer needed. Failure to do
    so will be ignored.

    :param apply_chown: Optional; change the owner of the directory.
    :type apply_chown: unicode
    :param apply_chmod: Optional; change the permissions of the directory.
    :type apply_chmod: unicode
    :param remove_using_sudo: Use sudo for removing the directory. ``None`` (default) means it is used depending on
     whether ``apply_chown`` has been set.
    :type remove_using_sudo: bool | NoneType
    :param remove_force: Force the removal.
    :type remove_force: bool
    :return: Path to the temporary directory.
    :rtype: unicode
    """
    path = get_remote_temp()
    try:
        if apply_chmod:
            run(chmod(apply_chmod, path))
        if apply_chown:
            if remove_using_sudo is None:
                remove_using_sudo = True
            sudo(chown(apply_chown, path))
        yield path
    finally:
        remove_ignore(path, use_sudo=remove_using_sudo, force=remove_force)


@documented_contextmanager
def local_temp_dir():
    """
    Creates a local temporary directory. The directory is removed when no longer needed. Failure to do
    so will be ignored.

    :return: Path to the temporary directory.
    :rtype: unicode
    """
    path = tempfile.mkdtemp()
    yield path
    shutil.rmtree(path, ignore_errors=True)


def extract_tar(filename, dest_path, **kwargs):
    """
    Extracts a TAR archive. All element names starting with ``/`` (indicating an absolute path) or that contain ``..``
    as references to a parent directory are not extracted.

    :param filename: Path to the tar file.
    :type filename: unicode
    :param dest_path: Destination path to extract the contents to.
    :type dest_path: unicode
    :param kwargs: Additional kwargs for opening the TAR file (:func:`tarfile.open`).
    """
    with tarfile.open(filename, 'r', **kwargs) as tf:
        safe_members = [name for name in tf.getmembers() if _safe_name(name)]
        if safe_members:
            tf.extractall(dest_path, safe_members)



from distutils.spawn import find_executable
import os
from setuptools import setup, find_packages

from dockerfabric import __version__


def include_readme():
    try:
        import pandoc
    except ImportError:
        return ''
    pandoc.core.PANDOC_PATH = find_executable('pandoc')
    readme_file = os.path.join(os.path.dirname(__file__), 'README.md')
    doc = pandoc.Document()
    with open(readme_file, 'r') as rf:
        doc.markdown = rf.read()
        return doc.rst


setup(
    name='docker-fabric',
    version=__version__,
    packages=find_packages(),
    install_requires=['six', 'Fabric>=1.8.0', 'docker-py>=0.5.0', 'docker-map>=0.7.0'],
    extras_require={
        'yaml': ['PyYAML'],
    },
    license='MIT',
    author='Matthias Erll',
    author_email='matthias@erll.de',
    url='https://github.com/merll/docker-fabric',
    description='Build Docker images, and run Docker containers in Fabric.',
    long_description=include_readme(),
    platforms=['OS Independent'],
    keywords=['docker', 'fabric'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Build Tools',
        'Topic :: System :: Software Distribution',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
    ],
    include_package_data=True,
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from fabric.api import env

from .apiclient import DockerFabricApiConnections, ContainerApiFabricClient, DockerClientConfiguration
from .cli import DockerCliConnections, ContainerCliFabricClient, DockerCliConfig

CLIENT_API = 'API'
CLIENT_CLI = 'CLI'

docker_api = DockerFabricApiConnections().get_connection
docker_cli = DockerCliConnections().get_connection


def docker_fabric(*args, **kwargs):
    """
    :param args: Positional arguments to Docker client.
    :param kwargs: Keyword arguments to Docker client.
    :return: Docker client.
    :rtype: dockerfabric.apiclient.DockerFabricClient | dockerfabric.cli.DockerCliClient
    """
    ci = kwargs.get('client_implementation') or env.get('docker_fabric_implementation') or CLIENT_API
    if ci == CLIENT_API:
        return docker_api(*args, **kwargs)
    elif ci == CLIENT_CLI:
        return docker_cli(*args, **kwargs)
    raise ValueError("Invalid client implementation.", ci)


def container_fabric(container_maps=None, docker_client=None, clients=None, client_implementation=None):
    """
    :param container_maps: Container map or a tuple / list thereof.
    :type container_maps: list[dockermap.map.config.main.ContainerMap] | dockermap.map.config.main.ContainerMap
    :param docker_client: Default Docker client instance.
    :type docker_client: dockerfabric.base.FabricClientConfiguration or docker.docker.Client
    :param clients: Optional dictionary of Docker client configuration objects.
    :type clients: dict[unicode | str, dockerfabric.base.FabricClientConfiguration]
    :param client_implementation: Client implementation to use (API or CLI).
    :type client_implementation: unicode | str
    :return: Container mapping client.
    :rtype: dockerfabric.base.FabricContainerClient
    """
    ci = client_implementation or env.get('docker_fabric_implementation') or CLIENT_API
    if ci == CLIENT_API:
        return ContainerApiFabricClient(container_maps, docker_client, clients)
    elif ci == CLIENT_CLI:
        return ContainerCliFabricClient(container_maps, docker_client, clients)
    raise ValueError("Invalid client implementation.", ci)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import ctypes
import logging
import multiprocessing

from dockermap.api import MappingDockerClient, ClientConfiguration
from fabric.api import env, settings

log = logging.getLogger(__name__)
port_offset = multiprocessing.Value(ctypes.c_ulong)


def _get_default_config(client_configs):
    clients = client_configs or env.get('docker_clients')
    host_string = env.get('host_string')
    if not host_string or not clients:
        return None
    for c in clients.values():
        host = c.get('fabric_host')
        if host == host_string:
            return c
    return None


class ConnectionDict(dict):
    def get_or_create_connection(self, key, d, *args, **kwargs):
        e = self.get(key)
        if e is None:
            log.debug("Creating new %s connection for key %s with args: %s, kwargs: %s",
                      self.__class__.__name__, key, args, kwargs)
            self[key] = e = d(*args, **kwargs)
        return e


class DockerConnectionDict(ConnectionDict):
    """
    Cache for connections to Docker clients.
    """
    configuration_class = None

    def get_connection(self, *args, **kwargs):
        """
        Create a new connection, or return an existing one from the cache. Uses Fabric's current ``env.host_string``
        and the URL to the Docker service.

        :param args: Additional arguments for the client constructor, if a new client has to be instantiated.
        :param kwargs: Additional keyword args for the client constructor, if a new client has to be instantiated.
        """
        key = env.get('host_string'), kwargs.get('base_url', env.get('docker_base_url'))
        default_config = _get_default_config(None)
        if default_config:
            if key not in self:
                self[key] = default_config
            return default_config.get_client()
        config = self.get_or_create_connection(key, self.configuration_class, *args, **kwargs)
        return config.get_client()


class FabricClientConfiguration(ClientConfiguration):
    def get_client(self):
        if 'fabric_host' in self:
            with settings(host_string=self.fabric_host):
                return super(FabricClientConfiguration, self).get_client()
        return super(FabricClientConfiguration, self).get_client()


class FabricContainerClient(MappingDockerClient):
    """
    Convenience class for using a :class:`~dockermap.map.config.main.ContainerMap` on a :class:`DockerFabricClient`.

    :param container_maps: Container map or a tuple / list thereof.
    :type container_maps: list[dockermap.map.config.main.ContainerMap] | dockermap.map.config.main.ContainerMap
    :param docker_client: Default Docker client instance.
    :type docker_client: FabricClientConfiguration
    :param clients: Optional dictionary of Docker client configuration objects.
    :type clients: dict[unicode | str, FabricClientConfiguration]
    """
    def __init__(self, container_maps=None, docker_client=None, clients=None):
        all_maps = container_maps or env.get('docker_maps', ())
        if not isinstance(all_maps, (list, tuple)):
            env_maps = all_maps,
        else:
            env_maps = all_maps
        all_configs = clients or env.get('docker_clients', dict())
        current_clients = dict()

        default_client = docker_client or _get_default_config(all_configs)
        for c_map in env_maps:
            map_clients = set(c_map.clients or ())
            for config_name, c_config in c_map:
                if c_config.clients:
                    map_clients.update(c_config.clients)
            for map_client in map_clients:
                if map_client not in current_clients:
                    client_config = all_configs.get(map_client)
                    if not client_config:
                        raise ValueError("Client '{0}' used in map '{1}' not configured.".format(map_client, c_map.name))
                    client_host = client_config.get('fabric_host')
                    if not client_host:
                        raise ValueError("Client '{0}' is configured, but has no 'fabric_host' definition.".format(map_client))
                    current_clients[map_client] = client_config

        if not (default_client or clients):
            default_client = self.configuration_class()
        super(FabricContainerClient, self).__init__(container_maps=all_maps, docker_client=default_client,
                                                    clients=current_clients)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def get_local_port(init_port):
    with port_offset.get_lock():
        current_offset = port_offset.value
        port_offset.value += 1
    return int(init_port) + current_offset



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from itertools import repeat
import re

from fabric.utils import error
from .output import stdout_result


IP4_PATTERN = re.compile(r'\s+inet addr:\s*((?:\d{1,3}\.){3}\d{1,3})')
IP6_PATTERN = re.compile(r'\s+inet6 addr:\s*((?:[a-f0-9]{1,4}::?){1,7}[a-f0-9]{1,4})')


def _get_address(interface_name, pattern):
    out = stdout_result('ifconfig {0}'.format(interface_name), (1,), shell=False, quiet=True)
    if not out:
        error("Network interface {0} not found.".format(interface_name))
    match = pattern.search(out)
    if match:
        return match.group(1)
    return None


def _expand_groups(address):
    for group in address.split(':'):
        if group:
            yield group.zfill(4)
        else:
            zero_groups = 8 - address.count(':') if '::' in address else 0
            for z in repeat('0000', zero_groups):
                yield z


def get_ip4_address(interface_name):
    """
    Extracts the IPv4 address for a particular interface from `ifconfig`.

    :param interface_name: Name of the network interface (e.g. ``eth0``).
    :type interface_name: unicode
    :return: IPv4 address; ``None`` if the interface is present but no address could be extracted.
    :rtype: unicode
    """
    return _get_address(interface_name, IP4_PATTERN)


def get_ip6_address(interface_name, expand=False):
    """
    Extracts the IPv6 address for a particular interface from `ifconfig`.

    :param interface_name: Name of the network interface (e.g. ``eth0``).
    :type interface_name: unicode
    :param expand: If set to ``True``, an abbreviated address is expanded to the full address.
    :type expand: bool
    :return: IPv6 address; ``None`` if the interface is present but no address could be extracted.
    :rtype: unicode
    """
    address = _get_address(interface_name, IP6_PATTERN)
    if address and expand:
        return ':'.join(_expand_groups(address))
    return address
